@@ -148,18 +148,14 @@ impl InputHandler {
                     }
                 },
                 Ok((InputInstruction::StartedParsing, _error_context)) => {
-                    log::info!("sending done loading");
                     self.send_client_instructions
                         .send(ClientInstruction::StartedParsingStdinQuery)
                         .unwrap();
-                    log::info!("done sent done loading");
                 },
                 Ok((InputInstruction::DoneParsing, _error_context)) => {
-                    log::info!("sending done loading");
                     self.send_client_instructions
                         .send(ClientInstruction::DoneParsingStdinQuery)
                         .unwrap();
-                    log::info!("done sent done loading");
                 },
                 Err(err) => panic!("Encountered read error: {:?}", err),
             }
@@ -311,6 +307,7 @@ impl InputHandler {
             | Action::GoToPreviousTab
             | Action::CloseTab
             | Action::GoToTab(_)
+            | Action::GoToTabName(_, _)
             | Action::ToggleTab
             | Action::MoveFocusOrTab(_) => {
                 self.command_is_executing.blocking_input_thread();
