@@ -18,8 +18,9 @@ use zellij_utils::pane_size::Offset;
 use zellij_utils::{
     data::{InputMode, Palette, PaletteColor, Style},
     errors::prelude::*,
+    input::layout::Run,
+    pane_size::PaneGeom,
     pane_size::SizeInPixels,
-    pane_size::{Dimension, PaneGeom},
     position::Position,
     shared::make_terminal_title,
     vte,
@@ -109,7 +110,9 @@ pub struct TerminalPane {
     is_held: Option<(Option<i32>, IsFirstRun, RunCommand)>, // a "held" pane means that its command has either exited and the pane is waiting for a
     // possible user instruction to be re-run, or that the command has not yet been run
     banner: Option<String>, // a banner to be rendered inside this TerminalPane, used for panes
-                            // held on startup and can possibly be used to display some errors
+    // held on startup and can possibly be used to display some errors
+    pane_frame_color_override: Option<(PaletteColor, Option<String>)>,
+    invoked_with: Option<Run>,
 }
 
 impl Pane for TerminalPane {
@@ -164,6 +167,10 @@ impl Pane for TerminalPane {
     }
     fn cursor_coordinates(&self) -> Option<(usize, usize)> {
         // (x, y)
+        if self.get_content_rows() < 1 || self.get_content_columns() < 1 {
+            // do not render cursor if there's no room for it
+            return None;
+        }
         let Offset { top, left, .. } = self.content_offset;
         self.grid
             .cursor_coordinates()
@@ -282,6 +289,11 @@ impl Pane for TerminalPane {
         if self.should_render() {
             let content_x = self.get_content_x();
             let content_y = self.get_content_y();
+            let rows = self.get_content_rows();
+            let columns = self.get_content_columns();
+            if rows < 1 || columns < 1 {
+                return Ok(None);
+            }
             match self.grid.render(content_x, content_y, &self.style) {
                 Ok(rendered_assets) => {
                     self.set_should_render(false);
@@ -301,7 +313,13 @@ impl Pane for TerminalPane {
     ) -> Result<Option<(Vec<CharacterChunk>, Option<String>)>> {
         let err_context = || format!("failed to render frame for client {client_id}");
         // TODO: remove the cursor stuff from here
-        let pane_title = if self.pane_name.is_empty()
+        let pane_title = if let Some(text_color_override) = self
+            .pane_frame_color_override
+            .as_ref()
+            .and_then(|(_color, text)| text.as_ref())
+        {
+            text_color_override.into()
+        } else if self.pane_name.is_empty()
             && input_mode == InputMode::RenamePane
             && frame_params.is_main_client
         {
@@ -340,8 +358,9 @@ impl Pane for TerminalPane {
             self.pane_name.clone()
         };
 
+        let frame_geom = self.current_geom();
         let mut frame = PaneFrame::new(
-            self.current_geom().into(),
+            frame_geom.into(),
             self.grid.scrollback_position_and_length(),
             pane_title,
             frame_params,
@@ -353,6 +372,9 @@ impl Pane for TerminalPane {
                 frame.add_exit_status(exit_status.as_ref().copied());
             }
         }
+        if let Some((frame_color_override, _text)) = self.pane_frame_color_override.as_ref() {
+            frame.override_color(*frame_color_override);
+        }
 
         let res = match self.frame.get(&client_id) {
             // TODO: use and_then or something?
@@ -436,25 +458,25 @@ impl Pane for TerminalPane {
     }
     fn reduce_height(&mut self, percent: f64) {
         if let Some(p) = self.geom.rows.as_percent() {
-            self.geom.rows = Dimension::percent(p - percent);
+            self.geom.rows.set_percent(p - percent);
             self.set_should_render(true);
         }
     }
     fn increase_height(&mut self, percent: f64) {
         if let Some(p) = self.geom.rows.as_percent() {
-            self.geom.rows = Dimension::percent(p + percent);
+            self.geom.rows.set_percent(p + percent);
             self.set_should_render(true);
         }
     }
     fn reduce_width(&mut self, percent: f64) {
         if let Some(p) = self.geom.cols.as_percent() {
-            self.geom.cols = Dimension::percent(p - percent);
+            self.geom.cols.set_percent(p - percent);
             self.set_should_render(true);
         }
     }
     fn increase_width(&mut self, percent: f64) {
         if let Some(p) = self.geom.cols.as_percent() {
-            self.geom.cols = Dimension::percent(p + percent);
+            self.geom.cols.set_percent(p + percent);
             self.set_should_render(true);
         }
     }
@@ -669,6 +691,23 @@ impl Pane for TerminalPane {
         }
         self.set_should_render(true);
     }
+    fn add_red_pane_frame_color_override(&mut self, error_text: Option<String>) {
+        self.pane_frame_color_override = Some((self.style.colors.red, error_text));
+    }
+    fn clear_pane_frame_color_override(&mut self) {
+        self.pane_frame_color_override = None;
+    }
+    fn frame_color_override(&self) -> Option<PaletteColor> {
+        self.pane_frame_color_override
+            .as_ref()
+            .map(|(color, _text)| *color)
+    }
+    fn invoked_with(&self) -> &Option<Run> {
+        &self.invoked_with
+    }
+    fn set_title(&mut self, title: String) {
+        self.pane_title = title;
+    }
 }
 
 impl TerminalPane {
@@ -685,6 +724,7 @@ impl TerminalPane {
         terminal_emulator_colors: Rc<RefCell<Palette>>,
         terminal_emulator_color_codes: Rc<RefCell<HashMap<usize, String>>>,
         initial_pane_title: Option<String>,
+        invoked_with: Option<Run>,
     ) -> TerminalPane {
         let initial_pane_title =
             initial_pane_title.unwrap_or_else(|| format!("Pane #{}", pane_index));
@@ -717,6 +757,8 @@ impl TerminalPane {
             search_term: String::new(),
             is_held: None,
             banner: None,
+            pane_frame_color_override: None,
+            invoked_with,
         }
     }
     pub fn get_x(&self) -> usize {
@@ -746,7 +788,7 @@ impl TerminalPane {
     fn reflow_lines(&mut self) {
         let rows = self.get_content_rows();
         let cols = self.get_content_columns();
-        self.grid.change_size(rows, cols);
+        self.grid.force_change_size(rows, cols);
         if self.banner.is_some() {
             self.grid.reset_terminal_state();
             self.render_first_run_banner();
@@ -758,6 +800,10 @@ impl TerminalPane {
     }
     pub fn cursor_coordinates(&self) -> Option<(usize, usize)> {
         // (x, y)
+        if self.get_content_rows() < 1 || self.get_content_columns() < 1 {
+            // do not render cursor if there's no room for it
+            return None;
+        }
         self.grid.cursor_coordinates()
     }
     fn render_first_run_banner(&mut self) {
