@@ -1,3 +1,4 @@
+use std::collections::VecDeque;
 use std::sync::{Arc, RwLock};
 
 use crate::{
@@ -165,6 +166,12 @@ pub(crate) fn route_action(
                 .send_to_screen(screen_instr)
                 .with_context(err_context)?;
         },
+        Action::MovePaneBackwards => {
+            session
+                .senders
+                .send_to_screen(ScreenInstruction::MovePaneBackwards(client_id))
+                .with_context(err_context)?;
+        },
         Action::DumpScreen(val, full) => {
             session
                 .senders
@@ -207,6 +214,12 @@ pub(crate) fn route_action(
                 .send_to_screen(ScreenInstruction::ScrollToBottom(client_id))
                 .with_context(err_context)?;
         },
+        Action::ScrollToTop => {
+            session
+                .senders
+                .send_to_screen(ScreenInstruction::ScrollToTop(client_id))
+                .with_context(err_context)?;
+        },
         Action::PageScrollUp => {
             session
                 .senders
@@ -271,9 +284,9 @@ pub(crate) fn route_action(
                 .send_to_pty(pty_instr)
                 .with_context(err_context)?;
         },
-        Action::EditFile(path_to_file, line_number, split_direction, should_float) => {
+        Action::EditFile(path_to_file, line_number, cwd, split_direction, should_float) => {
             let title = format!("Editing: {}", path_to_file.display());
-            let open_file = TerminalAction::OpenFile(path_to_file, line_number);
+            let open_file = TerminalAction::OpenFile(path_to_file, line_number, cwd);
             let pty_instr = match (split_direction, should_float) {
                 (Some(Direction::Left), false) => {
                     PtyInstruction::SpawnTerminalVertically(Some(open_file), Some(title), client_id)
@@ -430,12 +443,27 @@ pub(crate) fn route_action(
                 .send_to_screen(ScreenInstruction::CloseFocusedPane(client_id))
                 .with_context(err_context)?;
         },
-        Action::NewTab(tab_layout, tab_name) => {
+        Action::NewTab(
+            tab_layout,
+            floating_panes_layout,
+            swap_tiled_layouts,
+            swap_floating_layouts,
+            tab_name,
+        ) => {
             let shell = session.default_shell.clone();
+            let swap_tiled_layouts =
+                swap_tiled_layouts.unwrap_or_else(|| session.layout.swap_tiled_layouts.clone());
+            let swap_floating_layouts = swap_floating_layouts
+                .unwrap_or_else(|| session.layout.swap_floating_layouts.clone());
             session
                 .senders
                 .send_to_screen(ScreenInstruction::NewTab(
-                    shell, tab_layout, tab_name, client_id,
+                    shell,
+                    tab_layout,
+                    floating_panes_layout,
+                    tab_name,
+                    (swap_tiled_layouts, swap_floating_layouts),
+                    client_id,
                 ))
                 .with_context(err_context)?;
         },
@@ -469,6 +497,21 @@ pub(crate) fn route_action(
                 .send_to_screen(ScreenInstruction::GoToTab(i, Some(client_id)))
                 .with_context(err_context)?;
         },
+        Action::GoToTabName(name, create) => {
+            let shell = session.default_shell.clone();
+            let swap_tiled_layouts = session.layout.swap_tiled_layouts.clone();
+            let swap_floating_layouts = session.layout.swap_floating_layouts.clone();
+            session
+                .senders
+                .send_to_screen(ScreenInstruction::GoToTabName(
+                    name,
+                    (swap_tiled_layouts, swap_floating_layouts),
+                    shell,
+                    create,
+                    Some(client_id),
+                ))
+                .with_context(err_context)?;
+        },
         Action::TabNameInput(c) => {
             session
                 .senders
@@ -606,6 +649,24 @@ pub(crate) fn route_action(
                 .with_context(err_context)?;
         },
         Action::ToggleMouseMode => {}, // Handled client side
+        Action::PreviousSwapLayout => {
+            session
+                .senders
+                .send_to_screen(ScreenInstruction::PreviousSwapLayout(client_id))
+                .with_context(err_context)?;
+        },
+        Action::NextSwapLayout => {
+            session
+                .senders
+                .send_to_screen(ScreenInstruction::NextSwapLayout(client_id))
+                .with_context(err_context)?;
+        },
+        Action::QueryTabNames => {
+            session
+                .senders
+                .send_to_screen(ScreenInstruction::QueryTabNames(client_id))
+                .with_context(err_context)?;
+        },
     }
     Ok(should_break)
 }
@@ -614,14 +675,13 @@ pub(crate) fn route_action(
 macro_rules! send_to_screen_or_retry_queue {
     ($rlocked_sessions:expr, $message:expr, $instruction: expr, $retry_queue:expr) => {{
         match $rlocked_sessions.as_ref() {
-            Some(session_metadata) => {
-                session_metadata.senders.send_to_screen($message).unwrap();
-            },
+            Some(session_metadata) => session_metadata.senders.send_to_screen($message),
             None => {
                 log::warn!("Server not ready, trying to place instruction in retry queue...");
                 if let Some(retry_queue) = $retry_queue.as_mut() {
-                    retry_queue.push($instruction);
+                    retry_queue.push_back($instruction);
                 }
+                Ok(())
             },
         }
     }};
@@ -635,15 +695,17 @@ pub(crate) fn route_thread_main(
     mut receiver: IpcReceiverWithContext<ClientToServerMsg>,
     client_id: ClientId,
 ) -> Result<()> {
-    let mut retry_queue = vec![];
+    let mut retry_queue = VecDeque::new();
     let err_context = || format!("failed to handle instruction for client {client_id}");
     'route_loop: loop {
         match receiver.recv() {
             Some((instruction, err_ctx)) => {
                 err_ctx.update_thread_ctx();
-                let rlocked_sessions = session_data.read().unwrap();
+                let rlocked_sessions = session_data.read().to_anyhow().with_context(err_context)?;
                 let handle_instruction = |instruction: ClientToServerMsg,
-                                          mut retry_queue: Option<&mut Vec<ClientToServerMsg>>|
+                                          mut retry_queue: Option<
+                    &mut VecDeque<ClientToServerMsg>,
+                >|
                  -> Result<bool> {
                     let mut should_break = false;
                     match instruction {
@@ -675,18 +737,24 @@ pub(crate) fn route_thread_main(
                         ClientToServerMsg::TerminalResize(new_size) => {
                             session_state
                                 .write()
-                                .unwrap()
+                                .to_anyhow()
+                                .with_context(err_context)?
                                 .set_client_size(client_id, new_size);
-                            let min_size = session_state
+                            session_state
                                 .read()
-                                .unwrap()
-                                .min_client_terminal_size()
-                                .with_context(err_context)?;
-                            rlocked_sessions
-                                .as_ref()
-                                .unwrap()
-                                .senders
-                                .send_to_screen(ScreenInstruction::TerminalResize(min_size))
+                                .to_anyhow()
+                                .and_then(|state| {
+                                    state.min_client_terminal_size().ok_or(anyhow!(
+                                        "failed to determine minimal client terminal size"
+                                    ))
+                                })
+                                .and_then(|min_size| {
+                                    rlocked_sessions
+                                        .as_ref()
+                                        .context("couldn't get reference to read-locked session")?
+                                        .senders
+                                        .send_to_screen(ScreenInstruction::TerminalResize(min_size))
+                                })
                                 .with_context(err_context)?;
                         },
                         ClientToServerMsg::TerminalPixelDimensions(pixel_dimensions) => {
@@ -695,7 +763,8 @@ pub(crate) fn route_thread_main(
                                 ScreenInstruction::TerminalPixelDimensions(pixel_dimensions),
                                 instruction,
                                 retry_queue
-                            );
+                            )
+                            .with_context(err_context)?;
                         },
                         ClientToServerMsg::BackgroundColor(ref background_color_instruction) => {
                             send_to_screen_or_retry_queue!(
@@ -705,7 +774,8 @@ pub(crate) fn route_thread_main(
                                 ),
                                 instruction,
                                 retry_queue
-                            );
+                            )
+                            .with_context(err_context)?;
                         },
                         ClientToServerMsg::ForegroundColor(ref foreground_color_instruction) => {
                             send_to_screen_or_retry_queue!(
@@ -715,7 +785,8 @@ pub(crate) fn route_thread_main(
                                 ),
                                 instruction,
                                 retry_queue
-                            );
+                            )
+                            .with_context(err_context)?;
                         },
                         ClientToServerMsg::ColorRegisters(ref color_registers) => {
                             send_to_screen_or_retry_queue!(
@@ -723,7 +794,8 @@ pub(crate) fn route_thread_main(
                                 ScreenInstruction::TerminalColorRegisters(color_registers.clone()),
                                 instruction,
                                 retry_queue
-                            );
+                            )
+                            .with_context(err_context)?;
                         },
                         ClientToServerMsg::NewClient(
                             client_attributes,
@@ -776,7 +848,7 @@ pub(crate) fn route_thread_main(
                     }
                     Ok(should_break)
                 };
-                for instruction_to_retry in retry_queue.drain(..) {
+                while let Some(instruction_to_retry) = retry_queue.pop_front() {
                     log::warn!("Server ready, retrying sending instruction.");
                     let should_break = handle_instruction(instruction_to_retry, None)?;
                     if should_break {
