@@ -1,12 +1,15 @@
 //! Definition of the actions that can be bound to keys.
 
 use super::command::RunCommandAction;
-use super::layout::{Layout, PaneLayout};
+use super::layout::{
+    FloatingPaneLayout, Layout, SwapFloatingLayout, SwapTiledLayout, TiledPaneLayout,
+};
 use crate::cli::CliAction;
 use crate::data::InputMode;
 use crate::data::{Direction, Resize};
-use crate::input::config::{ConfigError, KdlError};
+use crate::input::config::{Config, ConfigError, KdlError};
 use crate::input::options::OnForceClose;
+use crate::setup::{find_default_config_dir, get_layout_dir};
 use miette::{NamedSource, Report};
 use serde::{Deserialize, Serialize};
 
@@ -115,6 +118,7 @@ pub enum Action {
     /// If there is no pane in the direction, move to previous/next Tab.
     MoveFocusOrTab(Direction),
     MovePane(Option<Direction>),
+    MovePaneBackwards,
     /// Dumps the screen to a file
     DumpScreen(String, bool),
     /// Scroll up in focus pane.
@@ -128,6 +132,8 @@ pub enum Action {
     ScrollDownAt(Position),
     /// Scroll down to bottom in focus pane.
     ScrollToBottom,
+    /// Scroll up to top in focus pane.
+    ScrollToTop,
     /// Scroll up one page in focus pane.
     PageScrollUp,
     /// Scroll down one page in focus pane.
@@ -146,7 +152,13 @@ pub enum Action {
     /// If no direction is specified, will try to use the biggest available space.
     NewPane(Option<Direction>, Option<String>), // String is an optional pane name
     /// Open the file in a new pane using the default editor
-    EditFile(PathBuf, Option<usize>, Option<Direction>, bool), // usize is an optional line number, bool is floating true/false
+    EditFile(
+        PathBuf,
+        Option<usize>,
+        Option<PathBuf>,
+        Option<Direction>,
+        bool,
+    ), // usize is an optional line number, Option<PathBuf> is an optional cwd, bool is floating true/false
     /// Open a new floating pane
     NewFloatingPane(Option<RunCommandAction>, Option<String>), // String is an optional pane name
     /// Open a new tiled (embedded, non-floating) pane
@@ -162,7 +174,13 @@ pub enum Action {
     PaneNameInput(Vec<u8>),
     UndoRenamePane,
     /// Create a new tab, optionally with a specified tab layout.
-    NewTab(Option<PaneLayout>, Option<String>), // the String is the tab name
+    NewTab(
+        Option<TiledPaneLayout>,
+        Vec<FloatingPaneLayout>,
+        Option<Vec<SwapTiledLayout>>,
+        Option<Vec<SwapFloatingLayout>>,
+        Option<String>,
+    ), // the String is the tab name
     /// Do nothing.
     NoOp,
     /// Go to the next tab.
@@ -172,6 +190,7 @@ pub enum Action {
     /// Close the current tab.
     CloseTab,
     GoToTab(u32),
+    GoToTabName(String, bool),
     ToggleTab,
     TabNameInput(Vec<u8>),
     UndoRenameTab,
@@ -202,12 +221,25 @@ pub enum Action {
     /// Toggle case sensitivity of search
     SearchToggleOption(SearchOption),
     ToggleMouseMode,
+    PreviousSwapLayout,
+    NextSwapLayout,
+    /// Query all tab names
+    QueryTabNames,
 }
 
 impl Action {
+    /// Checks that two Action are match except their mutable attributes.
+    pub fn shallow_eq(&self, other_action: &Action) -> bool {
+        match (self, other_action) {
+            (Action::NewTab(..), Action::NewTab(..)) => true,
+            _ => self == other_action,
+        }
+    }
+
     pub fn actions_from_cli(
         cli_action: CliAction,
         get_current_dir: Box<dyn Fn() -> PathBuf>,
+        config: Option<Config>,
     ) -> Result<Vec<Action>, String> {
         match cli_action {
             CliAction::Write { bytes } => Ok(vec![Action::Write(bytes)]),
@@ -217,7 +249,8 @@ impl Action {
             CliAction::FocusPreviousPane => Ok(vec![Action::FocusPreviousPane]),
             CliAction::MoveFocus { direction } => Ok(vec![Action::MoveFocus(direction)]),
             CliAction::MoveFocusOrTab { direction } => Ok(vec![Action::MoveFocusOrTab(direction)]),
-            CliAction::MovePane { direction } => Ok(vec![Action::MovePane(Some(direction))]),
+            CliAction::MovePane { direction } => Ok(vec![Action::MovePane(direction)]),
+            CliAction::MovePaneBackwards => Ok(vec![Action::MovePaneBackwards]),
             CliAction::DumpScreen { path, full } => Ok(vec![Action::DumpScreen(
                 path.as_os_str().to_string_lossy().into(),
                 full,
@@ -226,6 +259,7 @@ impl Action {
             CliAction::ScrollUp => Ok(vec![Action::ScrollUp]),
             CliAction::ScrollDown => Ok(vec![Action::ScrollDown]),
             CliAction::ScrollToBottom => Ok(vec![Action::ScrollToBottom]),
+            CliAction::ScrollToTop => Ok(vec![Action::ScrollToTop]),
             CliAction::PageScrollUp => Ok(vec![Action::PageScrollUp]),
             CliAction::PageScrollDown => Ok(vec![Action::PageScrollDown]),
             CliAction::HalfPageScrollUp => Ok(vec![Action::HalfPageScrollUp]),
@@ -292,13 +326,14 @@ impl Action {
                     .map(|cwd| current_dir.join(cwd))
                     .or_else(|| Some(current_dir));
                 if file.is_relative() {
-                    if let Some(cwd) = cwd {
+                    if let Some(cwd) = cwd.as_ref() {
                         file = cwd.join(file);
                     }
                 }
                 Ok(vec![Action::EditFile(
                     file,
                     line_number,
+                    cwd,
                     direction,
                     floating,
                 )])
@@ -318,21 +353,30 @@ impl Action {
             CliAction::GoToPreviousTab => Ok(vec![Action::GoToPreviousTab]),
             CliAction::CloseTab => Ok(vec![Action::CloseTab]),
             CliAction::GoToTab { index } => Ok(vec![Action::GoToTab(index)]),
+            CliAction::GoToTabName { name, create } => Ok(vec![Action::GoToTabName(name, create)]),
             CliAction::RenameTab { name } => Ok(vec![
                 Action::TabNameInput(vec![0]),
                 Action::TabNameInput(name.as_bytes().to_vec()),
             ]),
             CliAction::UndoRenameTab => Ok(vec![Action::UndoRenameTab]),
-            CliAction::NewTab { name, layout, cwd } => {
+            CliAction::NewTab {
+                name,
+                layout,
+                layout_dir,
+                cwd,
+            } => {
                 let current_dir = get_current_dir();
                 let cwd = cwd
                     .map(|cwd| current_dir.join(cwd))
                     .or_else(|| Some(current_dir));
                 if let Some(layout_path) = layout {
-                    let (path_to_raw_layout, raw_layout) =
-                        Layout::stringified_from_path_or_default(Some(&layout_path), None)
+                    let layout_dir = layout_dir
+                        .or_else(|| config.and_then(|c| c.options.layout_dir))
+                        .or_else(|| get_layout_dir(find_default_config_dir()));
+                    let (path_to_raw_layout, raw_layout, swap_layouts) =
+                        Layout::stringified_from_path_or_default(Some(&layout_path), layout_dir)
                             .map_err(|e| format!("Failed to load layout: {}", e))?;
-                    let layout = Layout::from_str(&raw_layout, path_to_raw_layout, cwd).map_err(|e| {
+                    let layout = Layout::from_str(&raw_layout, path_to_raw_layout, swap_layouts.as_ref().map(|(f, p)| (f.as_str(), p.as_str())), cwd).map_err(|e| {
                         let stringified_error = match e {
                             ConfigError::KdlError(kdl_error) => {
                                 let error = kdl_error.add_src(layout_path.as_path().as_os_str().to_string_lossy().to_string(), String::from(raw_layout));
@@ -368,17 +412,53 @@ impl Action {
                     if tabs.len() > 1 {
                         return Err(format!("Tab layout cannot itself have tabs"));
                     } else if !tabs.is_empty() {
-                        let (tab_name, layout) = tabs.drain(..).next().unwrap();
+                        let swap_tiled_layouts = if layout.swap_tiled_layouts.is_empty() {
+                            None
+                        } else {
+                            Some(layout.swap_tiled_layouts.clone())
+                        };
+                        let swap_floating_layouts = if layout.swap_floating_layouts.is_empty() {
+                            None
+                        } else {
+                            Some(layout.swap_floating_layouts.clone())
+                        };
+                        let (tab_name, layout, floating_panes_layout) =
+                            tabs.drain(..).next().unwrap();
                         let name = tab_name.or(name);
-                        Ok(vec![Action::NewTab(Some(layout), name)])
+                        Ok(vec![Action::NewTab(
+                            Some(layout),
+                            floating_panes_layout,
+                            swap_tiled_layouts,
+                            swap_floating_layouts,
+                            name,
+                        )])
                     } else {
-                        let layout = layout.new_tab();
-                        Ok(vec![Action::NewTab(Some(layout), name)])
+                        let swap_tiled_layouts = if layout.swap_tiled_layouts.is_empty() {
+                            None
+                        } else {
+                            Some(layout.swap_tiled_layouts.clone())
+                        };
+                        let swap_floating_layouts = if layout.swap_floating_layouts.is_empty() {
+                            None
+                        } else {
+                            Some(layout.swap_floating_layouts.clone())
+                        };
+                        let (layout, floating_panes_layout) = layout.new_tab();
+                        Ok(vec![Action::NewTab(
+                            Some(layout),
+                            floating_panes_layout,
+                            swap_tiled_layouts,
+                            swap_floating_layouts,
+                            name,
+                        )])
                     }
                 } else {
-                    Ok(vec![Action::NewTab(None, name)])
+                    Ok(vec![Action::NewTab(None, vec![], None, None, name)])
                 }
             },
+            CliAction::PreviousSwapLayout => Ok(vec![Action::PreviousSwapLayout]),
+            CliAction::NextSwapLayout => Ok(vec![Action::NextSwapLayout]),
+            CliAction::QueryTabNames => Ok(vec![Action::QueryTabNames]),
         }
     }
 }
